from pathlib import Path
from typing import Optional, Protocol

from attr import attrs
from pydantic import BaseModel, FilePath

from foundry.core.graphics_page import CHR_ROM_OFFSET, CHR_ROM_SEGMENT_SIZE
from foundry.game.File import ROM


class GraphicsPageProtocol(Protocol):
    """
    A representation of a single page of graphics inside the ROM.

    Attributes
    ----------
    index: int
        The index of the graphical page into the ROM.
    """

    index: int

    def __bytes__(self) -> bytes:
        ...


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class GraphicsPage:
    """
    A representation of a single page of graphics inside the ROM, that uses ``attrs`` to create a
    basic implementation.

    Attributes
    ----------
    index: int
        The index of the graphical page into the ROM.
    path: Optional[Path]
        The path to the file containing the bytes of the graphics page or ROM if None.
    """

    index: int
    path: Optional[Path] = None

    @property
    def offset(self) -> int:
        return CHR_ROM_OFFSET + self.index * CHR_ROM_SEGMENT_SIZE

    def __bytes__(self) -> bytes:
        if self.path is None:
            return bytes(ROM().bulk_read(CHR_ROM_SEGMENT_SIZE, self.offset))
        with open(self.path, "rb") as f:
            return f.read()[CHR_ROM_SEGMENT_SIZE * self.offset : CHR_ROM_SEGMENT_SIZE * (self.offset + 1)]


class PydanticGraphicsPage(BaseModel):
    """
    A JSON model of a generic GraphicsPage through Pydantic

    Attributes
    ----------
    index: int
        The index of the graphical page into the ROM.
    path: Optional[FilePath]
        The path to the file containing the bytes of the graphics page or ROM if None.
    """

    index: int
    path: Optional[FilePath]

    def to_graphics_page(self) -> GraphicsPageProtocol:
        return GraphicsPage(self.index, self.path)



from PySide6.QtGui import QColor

from foundry import data_dir, root_dir
from foundry.smb3parse.constants import BASE_OFFSET, Palette_By_Tileset, PalSet_Maps

MAP_PALETTE_ADDRESS = PalSet_Maps
PRG_SIZE = 0x2000
PALETTE_PRG_NO = 22
PALETTE_BASE_ADDRESS = BASE_OFFSET + PALETTE_PRG_NO * PRG_SIZE
PALETTE_OFFSET_LIST = Palette_By_Tileset
PALETTE_OFFSET_SIZE = 2  # bytes
PALETTE_GROUPS_PER_OBJECT_SET = 8
ENEMY_PALETTE_GROUPS_PER_OBJECT_SET = 4
PALETTES_PER_PALETTES_GROUP = 4
COLORS_PER_PALETTE = 4
COLOR_SIZE = 1  # byte
PALETTE_DATA_SIZE = (
    (PALETTE_GROUPS_PER_OBJECT_SET + ENEMY_PALETTE_GROUPS_PER_OBJECT_SET)
    * PALETTES_PER_PALETTES_GROUP
    * COLORS_PER_PALETTE
)
COLOR_COUNT = 64
BYTES_IN_COLOR = 3 + 1  # bytes + separator
PALETTE_FILE_PATH = data_dir / "palette.json"
palette_file = root_dir.joinpath("data", "Default.pal")

PALETTE_FILE_COLOR_OFFSET = 0x18
offset = 0x18  # first color point
NESPalette: list[QColor] = []
with open(palette_file, "rb") as f:
    color_data = f.read()

for i in range(COLOR_COUNT):
    NESPalette.append(QColor(color_data[offset], color_data[offset + 1], color_data[offset + 2]))

    offset += BYTES_IN_COLOR



from typing import Optional

from attr import attrs
from PySide6.QtCore import QPoint, QRect, Signal, SignalInstance
from PySide6.QtGui import (
    QBrush,
    QCloseEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
)
from PySide6.QtWidgets import QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.graphics_set.GraphicsSet import GraphicsSet
from foundry.core.palette.PaletteGroup import MutablePaletteGroup
from foundry.core.point.Point import Point
from foundry.game.gfx.drawable.Block import Block
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.PatternViewer import PatternViewerController as PatternViewer


@attrs(slots=True, auto_attribs=True)
class BlockEditorModel:
    tile_square_assembly: bytearray
    block_index: int
    graphics_set: GraphicsSet
    palette_group: MutablePaletteGroup
    palette_index: int


class BlockEditorController(CustomChildWindow):
    tile_square_assembly_changed: SignalInstance = Signal(bytearray)  # type: ignore
    block_index_changed: SignalInstance = Signal(int)  # type: ignore
    graphics_set_changed: SignalInstance = Signal(GraphicsSet)  # type: ignore
    palette_group_changed: SignalInstance = Signal(MutablePaletteGroup)  # type: ignore
    palette_index_changed: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        tsa_data: bytearray,
        block_index: int,
        graphics_set: GraphicsSet,
        palette_group: MutablePaletteGroup,
        palette_index: int,
    ):
        super().__init__(parent, "Block Editor")

        self.model = BlockEditorModel(tsa_data, block_index, graphics_set, palette_group, palette_index)
        self.view = BlockEditorView(self, tsa_data, block_index, graphics_set, palette_group, palette_index)
        self.pattern_viewer = None
        self._last_x = 0
        self._last_y = 0
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)

        self.view.pattern_selected.connect(lambda values: self._on_pattern_selected(*values))

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        def change_zoom(offset: int):
            self.view.zoom += offset

        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.addToolBar(self.toolbar)
        self.setStatusBar(QStatusBar(self))

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        if self.pattern_viewer is not None:
            self.pattern_viewer.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def tsa_data(self) -> bytearray:
        return self.model.tile_square_assembly

    @tsa_data.setter
    def tsa_data(self, value: bytearray):
        self.model.tile_square_assembly = value
        self.view.tsa_data = self.tsa_data
        self.tile_square_assembly_changed.emit(self.tsa_data)
        self.view.update()

    def silent_update_tsa_data(self, value: bytearray):
        self.model.tile_square_assembly = value
        self.view.tsa_data = self.tsa_data
        self.view.update()

    @property
    def block_index(self) -> int:
        return self.model.block_index

    @block_index.setter
    def block_index(self, value: int):
        self.model.block_index = value
        self.view.block_index = value
        self.graphics_set_changed.emit(self.block_index)
        self.view.update()
        self.palette_index = value // 0x40

    @property
    def graphics_set(self) -> GraphicsSet:
        return self.model.graphics_set

    @graphics_set.setter
    def graphics_set(self, value: GraphicsSet):
        self.model.graphics_set = value
        self.view.graphics_set = value
        self.graphics_set_changed.emit(self.graphics_set)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.graphics_set = value

    @property
    def palette_group(self) -> MutablePaletteGroup:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, value: MutablePaletteGroup):
        self.model.palette_group = value
        self.view.palette_group = value
        self.palette_group_changed.emit(self.palette_group)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.palette_group = value

    @property
    def palette_index(self) -> int:
        return self.model.palette_index

    @palette_index.setter
    def palette_index(self, value: int):
        self.model.palette_index = value
        self.view.palette_index = value
        self.palette_index_changed.emit(self.palette_index)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.palette_index = value

    def _on_pattern_selected(self, x: int, y: int):
        self._last_x = x
        self._last_y = y

        def remove_viewer(*_):
            self.pattern_viewer = None

        if self.pattern_viewer is None:
            self.pattern_viewer = PatternViewer(self, self.graphics_set, self.palette_group, self.block_index // 0x40)
            self.pattern_viewer.pattern_selected.connect(
                lambda ptn: self._on_block_pattern_change(self._last_x, self._last_y, ptn)
            )
            self.pattern_viewer.destroyed.connect(remove_viewer)
            self.pattern_viewer.show()
        else:
            self.pattern_viewer.palette_index = self.block_index // 0x40

    def _on_block_pattern_change(self, x: int, y: int, pattern: int):
        index = self.block_index + 0x100 * (x * 2 + y)
        tsa_data = self.tsa_data
        tsa_data[index] = pattern
        self.tsa_data = tsa_data


class BlockEditorView(QWidget):
    pattern_selected: SignalInstance = Signal(object)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        tsa_data: bytearray,
        block_index: int,
        graphics_set: GraphicsSet,
        palette_group: MutablePaletteGroup,
        palette_index: int,
        zoom: int = 16,
    ):
        super().__init__(parent)

        self.tsa_data = tsa_data
        self.block_index = block_index
        self.graphics_set = graphics_set
        self.palette_group = palette_group
        self.palette_index = palette_index
        self.zoom = zoom

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        self.setFixedSize(self.block_scale, self.block_scale)

    @property
    def block_scale(self) -> int:
        return Block.WIDTH * self.zoom

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = Point.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // (self.block_scale // 2), pos.y // (self.block_scale // 2)
        self.pattern_selected.emit((pos.x, pos.y))

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        bg_color = self.palette_group[0][0]
        painter.setBrush(QBrush(bg_color))
        painter.drawRect(QRect(QPoint(0, 0), self.size()))
        block = Block(
            self.block_index,
            tuple(tuple(c for c in pal) for pal in self.palette_group),
            self.graphics_set,
            self.tsa_data,
        )
        block.draw(painter, 0, 0, self.block_scale)



from functools import cached_property
from typing import Protocol

from attr import attrs, field, validators
from pydantic import BaseModel, validator
from PySide6.QtGui import QColor


class ColorProtocol(Protocol):
    red: int
    blue: int
    green: int
    alpha: int

    @property
    def qcolor(self) -> QColor:
        return QColor(self.red, self.green, self.blue, self.alpha)


def _check_in_color_range(inst, attr, value):
    if not 0 <= value <= 0xFF:
        raise ValueError(f"{value} is not inside the range 0-255")
    return value


@attrs(slots=True, frozen=True, eq=True, hash=True)
class Color:
    red: int = field(validator=[validators.instance_of(int), _check_in_color_range])
    blue: int = field(validator=[validators.instance_of(int), _check_in_color_range])
    green: int = field(validator=[validators.instance_of(int), _check_in_color_range])
    alpha: int = field(default=255, validator=[validators.instance_of(int), _check_in_color_range])

    @property
    def qcolor(self) -> QColor:
        return QColor(self.red, self.green, self.blue, self.alpha)


class PydanticColor(BaseModel):
    red: int
    green: int
    blue: int
    alpha: int = 255

    # Enable cached property to be ignored by Pydantic
    class Config:
        arbitrary_types_allowed = True
        keep_untouched = (cached_property,)

    @validator("red", "blue", "green", "alpha", allow_reuse=True)
    def check_color_range(cls, v):
        if not 0 <= v <= 0xFF:
            raise ValueError(f"{v} is not inside the range 0-255")
        return v

    @cached_property
    def color(self) -> ColorProtocol:
        return Color(self.red, self.green, self.blue, self.alpha)



from abc import ABC, abstractmethod
from typing import ClassVar, Protocol, Sequence, Type, TypeVar

from attr import attrs
from PySide6.QtGui import QColor

from foundry.core.palette import COLORS_PER_PALETTE, PALETTES_PER_PALETTES_GROUP
from foundry.core.palette.Palette import (
    AbstractPalette,
    HashablePaletteProtocol,
    MutablePalette,
    MutablePaletteProtocol,
    Palette,
    PaletteProtocol,
)
from foundry.core.palette.util import get_internal_palette_offset


class PaletteGroupProtocol(Protocol):
    """
    A representation of a group of palettes.
    """

    palettes: Sequence[PaletteProtocol]

    def __bytes__(self) -> bytes:
        ...

    def __getitem__(self, item: int) -> PaletteProtocol:
        ...

    @property
    def background_color(self) -> QColor:
        ...


class MutablePaletteGroupProtocol(PaletteGroupProtocol, Protocol):
    """
    A mutable representation of a group of palettes.
    """

    def __setitem__(self, key: int, value: PaletteProtocol):
        ...


class HashablePaletteGroupProtocol(PaletteGroupProtocol, Protocol):
    """
    A hashable and immutable representation of a group of palettes.
    """

    def __hash__(self) -> int:
        ...


_T = TypeVar("_T", bound="AbstractPaletteGroup")


class AbstractPaletteGroup(ABC):
    """
    A partial implementation of a palette group that implements the critical parts of a palette group
    independent of its mutability or ability to be hashed.
    """

    palettes: Sequence[PaletteProtocol]

    PALETTE_TYPE: ClassVar[Type[AbstractPalette]] = AbstractPalette

    def __bytes__(self) -> bytes:
        b = bytearray()
        for palette in self.palettes:
            b.extend(bytes(palette))
        return bytes(b)

    def __getitem__(self, item: int) -> PaletteProtocol:
        return self.palettes[item]

    @property
    def background_color(self) -> QColor:
        return self.palettes[0].colors[0]

    @classmethod
    @abstractmethod
    def from_values(cls: Type[_T], *values: PaletteProtocol) -> _T:
        """
        A generalized way to create itself from a series of values.

        Returns
        -------
        AbstractPaletteGroup
            The created palette group from the series.
        """
        ...

    @classmethod
    def from_palette_group(cls: Type[_T], palette_group: PaletteGroupProtocol) -> _T:
        """
        Generates a AbstractPaletteGroup of this type from another PaletteGroupProtocol.

        Parameters
        ----------
        palette_group : PaletteGroupProtocol
            The palette to be converted to this type.

        Returns
        -------
        AbstractPaletteGroup
            A palette group that is equal to the original palette group.
        """
        return cls.from_values(*palette_group.palettes)

    @classmethod
    def as_empty(cls: Type[_T]) -> _T:
        """
        Makes an empty palette group of default values.

        Returns
        -------
        AbstractPaletteGroup
            The palette group filled with default values.
        """
        return cls.from_values(*[cls.PALETTE_TYPE.as_empty() for _ in range(PALETTES_PER_PALETTES_GROUP)])

    @classmethod
    def from_rom(cls: Type[_T], address: int) -> _T:
        """
        Creates a palette group from an absolute address in ROM.

        Parameters
        ----------
        address : int
            The absolute address into the ROM.

        Returns
        -------
        AbstractPaletteGroup
            The palette group that represents the absolute address in ROM.
        """
        return cls.from_values(
            *[
                cls.PALETTE_TYPE.from_rom(address + offset)
                for offset in [COLORS_PER_PALETTE * i for i in range(PALETTES_PER_PALETTES_GROUP)]
            ]
        )

    @classmethod
    def from_tileset(cls, tileset: int, index: int):
        """
        Loads a palette group from a tileset with a given index.

        Parameters
        ----------
        tileset : int
            The index of the tileset.
        index : int
            The index of the palette group inside the tileset.

        Returns
        -------
        MutablePaletteGroup
            The PaletteGroup that represents the tileset's palette group at the provided offset.
        """
        offset = get_internal_palette_offset(tileset) + index * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE
        return cls.from_rom(offset)


_MT = TypeVar("_MT", bound="MutablePaletteGroup")


@attrs(slots=True, auto_attribs=True, eq=True)
class MutablePaletteGroup(AbstractPaletteGroup):
    """
    A concrete implementation of a mutable group of palettes.
    """

    palettes: list[MutablePaletteProtocol]

    PALETTE_TYPE: ClassVar[Type[MutablePalette]] = MutablePalette

    def __setitem__(self, key: int, value: MutablePaletteProtocol):
        self.palettes[key] = value

    @classmethod
    def from_values(cls: Type[_MT], *values: PaletteProtocol) -> _MT:
        return cls([cls.PALETTE_TYPE.from_palette(palette) for palette in values])


_PT = TypeVar("_PT", bound="PaletteGroup")


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class PaletteGroup(AbstractPaletteGroup):
    """
    A concrete implementation of a hashable and immutable group of palettes.
    """

    palettes: tuple[HashablePaletteProtocol]

    PALETTE_TYPE: ClassVar[Type[Palette]] = Palette

    @classmethod
    def from_values(cls: Type[_PT], *values: PaletteProtocol) -> _PT:
        return cls(tuple(cls.PALETTE_TYPE.from_palette(palette) for palette in values))



from PySide6.QtGui import QImage, QPainter


class Painter:
    """
    Creates a context manager for a QPainter to be applied on a QImage.

    If no context manger is provided for a QPainter, any error during its life-cycle could result in a crash.
    """

    def __init__(self, image: QImage):
        self.image = image

    def __enter__(self) -> QPainter:
        self.painter = QPainter(self.image)
        return self.painter

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.painter.end()



from typing import Optional

from attr import attrs
from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtGui import QCloseEvent, Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFormLayout,
    QGridLayout,
    QLayout,
    QStatusBar,
    QToolBar,
    QWidget,
)

from foundry import icon
from foundry.core.palette.Palette import MutablePalette, MutablePaletteProtocol
from foundry.core.palette.PaletteGroup import (
    MutablePaletteGroup,
    MutablePaletteGroupProtocol,
)
from foundry.core.player_animations import ANIMATION_WIDTH
from foundry.core.player_animations.PlayerAnimation import PlayerAnimation
from foundry.core.player_animations.util import (
    get_animations_palette_index,
    load_animations,
    load_player_animation_data,
    load_player_animations,
    load_power_up_offsets,
    load_power_up_palettes,
    save_player_animations_to_rom,
)
from foundry.core.point.Point import Point
from foundry.core.sprites import SPRITE_SIZE
from foundry.core.sprites.Sprite import Sprite, SpriteProtocol
from foundry.core.sprites.SpriteGroup import SpriteGroup, SpriteGroupProtocol
from foundry.core.UndoController import UndoController
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.PaletteEditorWidget import PaletteEditorWidget
from foundry.gui.PlayerFrameEditor import (
    PlayerFrameEditorController as PlayerFrameEditor,
)
from foundry.gui.SettingsDialog import SettingsDialog
from foundry.gui.Spinner import Spinner
from foundry.gui.SpriteViewerWidget import SpriteViewerWidget

POWERUPS = [
    ("Small Mario", 32, 53),
    ("Big Mario", 6, 48),
    ("Fire Mario", 16, 53),
    ("Raccoon Mario", 57, 53),
    ("Frog Mario", 56, 53),
    ("Tanooki Mario", 54, 53),
    ("Hammer Mario", 58, 53),
]


@attrs(slots=True, auto_attribs=True)
class PlayerViewerModel:
    is_mario: bool
    power_up: int
    power_up_offsets: list[int]
    palette_group: MutablePaletteGroupProtocol
    animations: list[PlayerAnimation]

    @classmethod
    def from_rom(
        cls,
        is_mario: bool = True,
        power_up: int = 0,
        power_up_offsets: Optional[list[int]] = None,
        palette_group: Optional[MutablePaletteGroupProtocol] = None,
        animations: Optional[list[PlayerAnimation]] = None,
    ):
        return cls(
            is_mario,
            power_up,
            power_up_offsets or load_power_up_offsets(),
            palette_group or load_power_up_palettes(),
            animations or load_player_animations(),
        )

    @classmethod
    def from_bytes(
        cls,
        is_mario: bool,
        power_up: int,
        power_up_offsets: bytes,
        palette_group: bytes,
        animations: bytes,
        page_offsets: bytes,
    ):
        return cls(
            is_mario,
            power_up,
            [i for i in power_up_offsets],
            MutablePaletteGroup(
                [MutablePalette([j for j in palette_group[i : (i + 4)]]) for i in range(len(palette_group) // 4)]
            ),
            load_animations(animations, page_offsets),
        )

    def to_bytes(self) -> tuple[bytes, bytes, bytes, bytes]:
        animation_bytes = bytearray()
        page_offset_bytes = bytearray()
        for animation in self.animations:
            ani_bytes, page_bytes = animation.to_bytes()
            ani_bytes = bytearray([i + 1 for i in ani_bytes])
            animation_bytes += ani_bytes
            page_offset_bytes += page_bytes

        return (
            bytes(o & 0xFF for o in self.power_up_offsets),
            bytes(self.palette_group),
            bytes(animation_bytes),
            bytes(page_offset_bytes),
        )


class PlayerViewerController(CustomChildWindow):
    power_up_changed: SignalInstance = Signal(int)  # type: ignore
    is_mario_changed: SignalInstance = Signal(bool)  # type: ignore
    palette_group_changed: SignalInstance = Signal(MutablePaletteGroupProtocol)  # type: ignore
    power_up_offsets_changed: SignalInstance = Signal(object)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        title: str = "Player Viewer",
        is_mario: bool = True,
        power_up: int = 0,
        power_up_offsets: Optional[list[int]] = None,
        palette_group: Optional[MutablePaletteGroupProtocol] = None,
        animations: Optional[list[PlayerAnimation]] = None,
    ):
        super().__init__(parent, title)

        self.model = PlayerViewerModel.from_rom(is_mario, power_up, power_up_offsets, palette_group, animations)
        self.view = PlayerViewerView(self, self.sprite_groups)
        self.undo_controller = UndoController(self.model.to_bytes())
        self.frame_editor: Optional[PlayerFrameEditor] = None
        self.frame_editor_index = 0
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)
        self.side_toolbar = QToolBar(self)

        self.toolbar.setMovable(False)
        self.side_toolbar.setMovable(False)

        self.view.mouse_moved_over_widget.connect(self._on_mouse_move)
        self.view.clicked.connect(self._show_player_frame_editor)

        self.undo_action = self.toolbar.addAction(icon("rotate-ccw.svg"), "Undo Action")
        self.redo_action = self.toolbar.addAction(icon("rotate-cw.svg"), "Redo Action")
        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        self.powerup_combo_box = QComboBox()
        for (name, x, y) in POWERUPS:
            powerup_icon = SettingsDialog._load_from_png(x, y)
            self.powerup_combo_box.addItem(powerup_icon, name)
        self.is_mario_check_box = QCheckBox()
        self.palette_editor = PaletteEditorWidget(self, self.palette)
        self.power_up_offset_spinner = Spinner(self, 0xFF)

        def change_zoom(offset: int):
            self.view.zoom += offset

        def undo(*_):
            self.model = PlayerViewerModel.from_bytes(self.is_mario, self.power_up, *self.undo_controller.undo())
            self._update()

        def redo(*_):
            self.model = PlayerViewerModel.from_bytes(self.is_mario, self.power_up, *self.undo_controller.redo())
            self._update()

        def set_power_up(value: int):
            self.power_up = value

        def set_is_mario(value: bool):
            self.is_mario = value

        def set_palette(value: MutablePaletteProtocol):
            self.palette = value

        def set_power_up_offset(value: int):
            self.power_up_offset = value

        self.undo_action.triggered.connect(undo)  # type: ignore
        self.undo_action.setEnabled(False)
        self.redo_action.triggered.connect(redo)  # type: ignore
        self.redo_action.setEnabled(False)
        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.powerup_combo_box.currentIndexChanged.connect(set_power_up)  # type: ignore
        self.powerup_combo_box.setCurrentIndex(power_up)

        self.is_mario_check_box.setChecked(is_mario)
        self.is_mario_check_box.toggled.connect(set_is_mario)  # type: ignore

        self.palette_editor.palette_changed.connect(set_palette)

        self.power_up_offset_spinner.valueChanged.connect(set_power_up_offset)  # type: ignore

        side_toolbar_layout = QFormLayout(self)
        side_toolbar_layout.addRow("Power Up", self.powerup_combo_box)
        side_toolbar_layout.addRow("Mario or Luigi", self.is_mario_check_box)
        side_toolbar_layout.addRow("palette Editor", self.palette_editor)
        side_toolbar_layout.addRow("Page Offset", self.power_up_offset_spinner)

        class LayoutWidget(QWidget):
            def __init__(self, parent, layout: QLayout):
                super().__init__(parent)
                self.setLayout(layout)

        self.side_toolbar.addWidget(LayoutWidget(self.side_toolbar, side_toolbar_layout))

        self.addToolBar(self.toolbar)
        self.addToolBar(Qt.RightToolBarArea, self.side_toolbar)

        self.setStatusBar(QStatusBar(self))
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        if self.frame_editor is not None:
            self.frame_editor.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def power_up(self) -> int:
        return self.model.power_up

    @power_up.setter
    def power_up(self, power_up: int):
        self.model.power_up = power_up
        self._update_palette_editor()
        self._update_view_animations()
        self.power_up_offset_spinner.setValue(self.power_up_offset)
        self.powerup_combo_box.setCurrentIndex(power_up)
        self.power_up_changed.emit(power_up)

    @property
    def is_mario(self) -> bool:
        return self.model.is_mario

    @is_mario.setter
    def is_mario(self, is_mario: bool):
        self.model.is_mario = is_mario
        self._update()
        self.is_mario_check_box.setChecked(is_mario)
        self.is_mario_changed.emit(is_mario)

    @property
    def power_up_offset(self) -> int:
        return self.power_up_offsets[self.power_up]

    @power_up_offset.setter
    def power_up_offset(self, offset: int):
        offsets = self.power_up_offsets
        offsets[self.power_up] = offset
        self.power_up_offsets = offsets

    @property
    def power_up_offsets(self) -> list[int]:
        return self.model.power_up_offsets

    @power_up_offsets.setter
    def power_up_offsets(self, offsets: list[int]):
        self.model.power_up_offsets = offsets
        self.undo_controller.do(self.model.to_bytes())
        self.update()
        self.power_up_offsets_changed.emit(offsets)

    @property
    def palette(self) -> MutablePaletteProtocol:
        return self.palette_group[get_animations_palette_index(self.is_mario, self.power_up)]

    @palette.setter
    def palette(self, palette: MutablePaletteProtocol):
        palette_group = self.palette_group
        palette_group[get_animations_palette_index(self.is_mario, self.power_up)] = palette
        self.palette_group = palette_group

    @property
    def palette_group(self) -> MutablePaletteGroupProtocol:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, palette_group: MutablePaletteGroupProtocol):
        self.model.palette_group = palette_group
        self.undo_controller.do(self.model.to_bytes())
        self._update()
        self.palette_group_changed.emit(palette_group)

    @property
    def sprite_groups(self) -> list[SpriteGroupProtocol]:
        sprite_groups: list[SpriteGroupProtocol] = []

        for animation in load_player_animation_data(
            self.model.animations,
            self.model.palette_group,
            self.model.is_mario,
            self.model.power_up,
            self.model.power_up_offsets,
        ):
            sprites: list[SpriteProtocol] = []

            for idx, sprite in enumerate(animation.frames):
                sprites.append(
                    Sprite(
                        Point(
                            (idx % ANIMATION_WIDTH) * SPRITE_SIZE.width,
                            (idx // ANIMATION_WIDTH) * SPRITE_SIZE.height,
                        ),
                        sprite,
                        animation.palette_index,
                        animation.horizontal_flip[idx],
                        do_not_render=sprite == 0xF0,
                    )
                )

            sprite_groups.append(SpriteGroup(Point(0, 0), sprites, animation.graphics_set, animation.palette_group))

        return sprite_groups

    def _update(self):
        self.undo_action.setEnabled(self.undo_controller.can_undo)
        self.redo_action.setEnabled(self.undo_controller.can_redo)
        save_player_animations_to_rom(*self.model.to_bytes())
        self._update_palette_editor()
        self._update_view_animations()
        if self.frame_editor is not None:
            self.frame_editor.is_mario = self.is_mario

    def _update_palette_editor(self):
        self.palette_editor._palette = self.palette
        self.palette_editor._update()

    def _update_view_animations(self):
        self.view.sprite_groups = self.sprite_groups
        self.view._update()

    def _show_player_frame_editor(self, index: int):
        self.frame_editor_index = index

        if self.frame_editor is None:
            self.frame_editor = PlayerFrameEditor(
                self, self.model.animations[index], self.power_up_offsets, self.palette_group, self.is_mario
            )

            def remove_frame_editor():
                self.frame_editor = None

            def change_frame_offset(offset: int):
                self.model.animations[self.frame_editor_index].offset = offset
                self.undo_controller.do(self.model.to_bytes())
                self._update()

            def change_frame_sprites(sprites: bytearray):
                self.model.animations[self.frame_editor_index].animations = sprites
                self.undo_controller.do(self.model.to_bytes())
                self._update()

            self.frame_editor.destroyed.connect(remove_frame_editor)
            self.frame_editor.offset_changed.connect(change_frame_offset)
            self.frame_editor.frames_changed.connect(change_frame_sprites)
            self.frame_editor.show()
        else:
            self.frame_editor.model.animation = self.model.animations[index]
            self.frame_editor._update_view_animations()

    def _on_mouse_move(self, x: int, y: int, index: int, hex_index: str):
        self.statusBar().showMessage(f"Row: {y}, Column: {x}, Index: {index} / {hex_index}")


class PlayerViewerView(QWidget):
    mouse_moved_over_widget: SignalInstance = Signal(int, int, int, str)  # type: ignore
    clicked: SignalInstance = Signal(int)  # type: ignore

    ANIMATIONS_PER_ROW = 9
    ANIMATIONS_PER_COLUMN = 9

    def __init__(
        self,
        parent: Optional[QWidget],
        sprite_groups: list[SpriteGroupProtocol],
        zoom: int = 2,
    ):
        super().__init__(parent)

        self.sprite_groups = sprite_groups
        self._zoom = zoom

        self.layout_ = QGridLayout(self)
        self.widgets: list[SpriteViewerWidget] = []

        for idx, sprite_group in enumerate(sprite_groups):
            widget = SpriteViewerWidget(self, sprite_group, zoom=self.zoom)
            widget.mouse_moved_over_widget.connect(
                lambda *_, idx=idx: self.mouse_moved_over_widget.emit(
                    idx % self.ANIMATIONS_PER_ROW,
                    idx // self.ANIMATIONS_PER_ROW,
                    idx,
                    hex(idx).upper().replace("X", "x"),
                )
            )
            widget.clicked.connect(lambda *_, idx=idx: self.clicked.emit(idx))
            self.widgets.append(widget)
            self.layout_.addWidget(widget, idx // self.ANIMATIONS_PER_ROW, idx % self.ANIMATIONS_PER_ROW)

        self.setLayout(self.layout_)

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        for widget in self.widgets:
            widget.zoom = value

    def _update(self):
        for (widget, sprite_group) in zip(self.widgets, self.sprite_groups):
            widget.sprite_group = sprite_group
            widget.update()



from PySide6.QtCore import QPoint, QSize

from foundry.core.graphics_set.GraphicsSet import GraphicsSet
from foundry.core.palette.PaletteGroup import MutablePaletteGroup
from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.MapObject import MapObject
from foundry.game.level.LevelLike import LevelLike
from foundry.smb3parse.levels.world_map import (
    WORLD_MAP_HEIGHT,
    WORLD_MAP_SCREEN_SIZE,
    WORLD_MAP_SCREEN_WIDTH,
)
from foundry.smb3parse.levels.world_map import WorldMap as _WorldMap
from foundry.smb3parse.objects.object_set import WORLD_MAP_OBJECT_SET

OVERWORLD_GRAPHIC_SET = 0


class WorldMap(LevelLike):
    def __init__(self, world_index):
        self._internal_world_map = _WorldMap.from_world_number(ROM(), world_index)

        super(WorldMap, self).__init__(0, self._internal_world_map.layout_address)

        self.name = f"World {world_index} - Overworld"

        self.graphics_set = GraphicsSet.from_tileset(OVERWORLD_GRAPHIC_SET)
        self.palette_group = MutablePaletteGroup.from_tileset(WORLD_MAP_OBJECT_SET, 0)

        self.object_set = WORLD_MAP_OBJECT_SET
        self.tsa_data = ROM.get_tsa_data(self.object_set)

        self.world = 0
        self.level_number = world_index

        self.objects = []

        self._load_objects()

        self._calc_size()

    def _load_objects(self):
        self.objects.clear()

        for index, world_position in enumerate(self._internal_world_map.gen_positions()):
            screen_offset = (index // WORLD_MAP_SCREEN_SIZE) * WORLD_MAP_SCREEN_WIDTH

            x = screen_offset + (index % WORLD_MAP_SCREEN_WIDTH)
            y = (index // WORLD_MAP_SCREEN_WIDTH) % WORLD_MAP_HEIGHT

            block = Block(
                world_position.tile(),
                tuple(tuple(c for c in pal) for pal in self.palette_group),
                self.graphics_set,
                self.tsa_data,
            )

            self.objects.append(MapObject(block, x, y))

        assert len(self.objects) % WORLD_MAP_HEIGHT == 0

    def _calc_size(self):
        self.width = len(self.objects) // WORLD_MAP_HEIGHT
        self.height = WORLD_MAP_HEIGHT

        self.size = self.width, self.height

    def add_object(self, obj, _):
        self.objects.append(obj)

        self.objects.sort(key=self._array_index)

    @property
    def q_size(self):
        return QSize(*self.size) * Block.SIDE_LENGTH

    @staticmethod
    def _array_index(obj):
        return obj.y_position * WORLD_MAP_SCREEN_WIDTH + obj.x_position

    def get_object_names(self):
        return [obj.name for obj in self.objects]

    def draw(self, dc, zoom, transparency=None, show_expansion=None):
        for obj in self.objects:
            obj.draw(dc, Block.SIDE_LENGTH * zoom, transparency)

    def index_of(self, obj):
        return self.objects.index(obj)

    def get_all_objects(self):
        return self.objects

    def object_at(self, x, y):
        point = QPoint(x, y)

        for obj in reversed(self.objects):
            if obj.rect.contains(point):
                return obj

        return None

    def to_bytes(self):
        return_array = bytearray(len(self.objects))

        for obj in self.objects:
            index = self._array_index(obj)

            return_array[index] = obj.to_bytes()

        return self.layout_address, return_array

    def from_bytes(self, data, _=None):
        offset, obj_bytes = data

        self.layout_address = offset
        self._load_objects(obj_bytes)

        self._calc_size()

    def get_object(self, index):
        return self.objects[index]

    def remove_object(self, obj):
        self.objects.remove(obj)

    def level_at_position(self, x: int, y: int):
        assert isinstance(x, int)
        assert isinstance(y, int)

        screen = x // WORLD_MAP_SCREEN_WIDTH + 1

        x %= WORLD_MAP_SCREEN_WIDTH

        return self._internal_world_map.level_for_position(screen, y, x)

    def level_name_at_position(self, x: int, y: int) -> str:
        screen = x // WORLD_MAP_SCREEN_WIDTH + 1

        x %= WORLD_MAP_SCREEN_WIDTH

        return self._internal_world_map.level_name_for_position(screen, y, x)



from typing import Union

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.ObjectLike import EXPANDS_HORIZ, EXPANDS_VERT


def increment_type(item: Union[LevelObject, EnemyObject]):
    change_type(item, True)


def decrement_type(item: Union[LevelObject, EnemyObject]):
    change_type(item, False)


def change_type(item: Union[LevelObject, EnemyObject], increment: bool):
    if isinstance(item, LevelObject):
        return change_level_object_type(item, increment)
    elif isinstance(item, EnemyObject):
        return change_enemy_object_type(item, increment)


def change_level_object_type(item: LevelObject, increment: bool):
    value = 1 if increment else -1

    new_type = item.obj_index + value

    if new_type < 0 and item.domain > 0:
        new_domain = item.domain - 1
        new_type = 0xF0
    elif new_type > 0xFF and item.domain < 7:
        new_domain = item.domain + 1
        new_type = 0x00
    else:
        new_type = min(0xFF, new_type)
        new_type = max(0, new_type)

        new_domain = item.domain

    item.data[0] &= 0b0001_1111
    item.data[0] |= new_domain << 5
    item.data[2] = new_type

    if item.is_4byte and item.size == 3:
        item.data.append(0)

    item.render()


def change_enemy_object_type(item: EnemyObject, increment: bool):
    item.enemy.type = max(0, min(0xFF, item.obj_index + (1 if increment else -1)))
    item.render()


def set_level_object_width(item: LevelObject, width: int, *, render: bool = True):
    if not item.horizontally_expands:
        return

    if item.primary_expansion() == EXPANDS_HORIZ:
        item.obj_index = (item.obj_index & 0xF0) + max(0, min(0x0F, width - item.position.x))
    else:
        if item.is_4byte:
            item.data[3] = max(0, min(0xFF, width - item.position.x))
        else:
            raise NotImplementedError(f"Resize is not possible for {item}")

    if render:
        item._render()


def set_level_object_height(item: LevelObject, height: int, *, render: bool = True):
    if not item.vertically_expands:
        return

    if item.primary_expansion() == EXPANDS_VERT:
        item.obj_index = (item.obj_index & 0xF0) + max(0, min(0x0F, height - item.position.y))
    else:
        if item.is_4byte:
            item.data[3] = max(0, min(0xFF, height - item.position.y))
        else:
            raise NotImplementedError(f"Resize is not possible for {item}")

    if render:
        item._render()


def resize_level_object(item: LevelObject, width_difference: int, height_difference: int):
    if width_difference:
        set_level_object_width(item, item.position.x + width_difference, render=False)

    if height_difference:
        set_level_object_height(item, item.position.y + height_difference, render=False)

    item._render()



from foundry.core.graphics_page.GraphicsPage import GraphicsPage, GraphicsPageProtocol
from foundry.game.File import ROM
from foundry.smb3parse.constants import Level_BG_Pages1, Level_BG_Pages2

WORLD_MAP = 0
SPADE_ROULETTE = 16
N_SPADE = 17
VS_2P = 18
HILLY = 3
CORRECTED_HILLY = 3
UNDERGROUND = 14
CORRECTED_UNDERGROUND = 19

BG_PAGE_COUNT = Level_BG_Pages2 - Level_BG_Pages1  # 23 in stock rom

GRAPHIC_SET_NAMES = [
    "Mario graphics (1)",
    "Plain",
    "Dungeon",
    "Underground (1)",
    "Sky",
    "Pipe/Water (1, Piranha Plant)",
    "Pipe/Water (2, Water)",
    "Mushroom house (1)",
    "Pipe/Water (3, Pipe)",
    "Desert",
    "Ship",
    "Giant",
    "Ice",
    "Clouds",
    "Underground (2)",
    "Spade bonus room",
    "Spade bonus",
    "Mushroom house (2)",
    "Pipe/Water (4)",
    "Hills",
    "Plain 2",
    "Tank",
    "Castle",
    "Mario graphics (2)",
    "Animated graphics (1)",
    "Animated graphics (2)",
    "Animated graphics (3)",
    "Animated graphics (4)",
    "Animated graphics (P-Switch)",
    "Game font/Course Clear graphics",
    "Animated graphics (5)",
    "Animated graphics (6)",
]


def get_graphics_pages_from_tileset(tileset: int) -> tuple[GraphicsPageProtocol, ...]:
    if tileset == WORLD_MAP:
        return (
            GraphicsPage(0x14),
            GraphicsPage(0x15),
            GraphicsPage(0x16),
            GraphicsPage(0x17),
            GraphicsPage(0x20),
            GraphicsPage(0x21),
            GraphicsPage(0x22),
            GraphicsPage(0x23),
        )
    if tileset not in range(BG_PAGE_COUNT):
        return (
            GraphicsPage(tileset),
            GraphicsPage(tileset + 1),
            GraphicsPage(tileset + 2),
            GraphicsPage(tileset + 3),
        )
    if tileset == HILLY:
        tileset = CORRECTED_HILLY
    if tileset == UNDERGROUND:
        tileset = CORRECTED_UNDERGROUND

    graphic_page_index_1 = ROM().bulk_read(BG_PAGE_COUNT, Level_BG_Pages1)
    graphic_page_index_2 = ROM().bulk_read(BG_PAGE_COUNT, Level_BG_Pages2)
    pages = [
        GraphicsPage(graphic_page_index_1[tileset]),
        GraphicsPage(graphic_page_index_1[tileset] + 1),
        GraphicsPage(graphic_page_index_2[tileset]),
        GraphicsPage(graphic_page_index_2[tileset] + 1),
    ]

    if tileset == SPADE_ROULETTE:
        pages.extend([GraphicsPage(0x20), GraphicsPage(0x21), GraphicsPage(0x22), GraphicsPage(0x23)])
    elif tileset == N_SPADE:
        pages.extend([GraphicsPage(0x28), GraphicsPage(0x29), GraphicsPage(0x5A), GraphicsPage(0x31)])
    elif tileset == VS_2P:
        pages.extend([GraphicsPage(0x04), GraphicsPage(0x05), GraphicsPage(0x06), GraphicsPage(0x07)])
    else:
        pages.extend([GraphicsPage(0x00), GraphicsPage(0x00), GraphicsPage(0x00), GraphicsPage(0x00)])

    return tuple(pages)



from abc import ABC, abstractmethod
from enum import Enum
from functools import cached_property
from typing import Optional, Protocol, Sequence, Type, TypeVar

from attr import attrs
from pydantic import BaseModel
from PySide6.QtGui import QColor

from foundry.core.palette import COLORS_PER_PALETTE
from foundry.core.palette.ColorPalette import (
    ColorPaletteCreator,
    ColorPaletteProtocol,
    PydanticColorPaletteProtocol,
)
from foundry.game.File import ROM


class PaletteProtocol(Protocol):
    color_indexes: Sequence[int]
    color_palette: ColorPaletteProtocol

    def __bytes__(self) -> bytes:
        ...

    def __getitem__(self, item: int) -> int:
        ...

    @property
    def colors(self) -> Sequence[QColor]:
        ...


class MutablePaletteProtocol(PaletteProtocol, Protocol):
    def __setitem__(self, key: int, value: int):
        ...


class HashablePaletteProtocol(PaletteProtocol, Protocol):
    def __hash__(self) -> int:
        ...


_T = TypeVar("_T", bound="AbstractPalette")


class AbstractPalette(ABC):
    """
    An abstract implementation of the primary methods of a Palette, regardless of color_indexes.
    """

    color_indexes: Sequence[int]
    color_palette: ColorPaletteProtocol

    def __bytes__(self) -> bytes:
        return bytes([i & 0xFF for i in self.color_indexes])

    def __getitem__(self, item: int) -> int:
        return self.color_indexes[item]

    @property
    def colors(self) -> Sequence[QColor]:
        return [
            self.color_palette.colors[index & len(self.color_palette.colors)].qcolor for index in self.color_indexes
        ]

    @classmethod
    @abstractmethod
    def from_values(
        cls: Type[_T], color_indexes: Sequence[int], color_palette: Optional[ColorPaletteProtocol] = None
    ) -> _T:
        """
        A generalized way to create itself from a series of values.

        Returns
        -------
        AbstractPalette
            The created palette from the series.
        """
        ...

    @classmethod
    def from_palette(cls: Type[_T], palette: PaletteProtocol) -> _T:
        """
        Generates a AbstractPalette of this type from another PaletteProtocol.

        Parameters
        ----------
        palette : PaletteProtocol
            The palette to be converted to this type.

        Returns
        -------
        AbstractPalette
            A palette that is equal to the original palette.
        """
        return cls.from_values(palette.color_indexes, palette.color_palette)

    @classmethod
    def as_empty(cls: Type[_T]) -> _T:
        """
        Makes an empty palette of default values.

        Returns
        -------
        AbstractPalette
            A palette filled with default values.
        """
        return cls.from_values([0, 0, 0, 0])

    @classmethod
    def from_rom(cls: Type[_T], address: int) -> _T:
        """
        Creates a palette from an absolute address in ROM.

        Parameters
        ----------
        address : int
            The absolute address into the ROM.

        Returns
        -------
        AbstractPalette
            The palette that represents the absolute address in ROM.
        """
        return cls.from_values(tuple(int(i) for i in ROM().read(address, COLORS_PER_PALETTE)))


_MT = TypeVar("_MT", bound="MutablePalette")


@attrs(slots=True, auto_attribs=True, eq=True)
class MutablePalette(AbstractPalette):
    color_indexes: list[int]
    color_palette: ColorPaletteProtocol = ColorPaletteCreator.as_default().color_palette

    def __setitem__(self, key: int, value: int):
        self.color_indexes[key] = value

    @classmethod
    def from_values(
        cls: Type[_MT], color_indexes: Sequence[int], color_palette: Optional[ColorPaletteProtocol] = None
    ) -> _MT:
        if color_palette is None:
            return cls(list(color_indexes))
        else:
            return cls(list(color_indexes), color_palette)


_PT = TypeVar("_PT", bound="Palette")


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class Palette(AbstractPalette):
    color_indexes: tuple[int, ...]
    color_palette: ColorPaletteProtocol = ColorPaletteCreator.as_default().color_palette

    @classmethod
    def from_values(
        cls: Type[_PT], color_indexes: Sequence[int], color_palette: Optional[ColorPaletteProtocol] = None
    ) -> _PT:
        if color_palette is None:
            return cls(tuple(color_indexes))
        else:
            return cls(tuple(color_indexes), color_palette)


class PaletteType(str, Enum):
    """
    A declaration of the palettes possible to be created through
    `JSON <https://en.wikipedia.org/wiki/JSON>`_ and
    `Pydantic <https://pydantic-docs.helpmanual.io/>`_.
    """

    from_colors = "COLORS"
    from_rom_address = "ROM ADDRESS"

    @classmethod
    def has_value(cls, value):
        """
        A convenience method to quickly determine if a value is a valid enumeration.

        Parameters
        ----------
        value : str
            The value to check against the enumeration.

        Returns
        -------
        bool
            If the value is inside the enumeration.
        """

        return value in cls._value2member_map_


class PydanticPalette(BaseModel):
    """
    A generic representation of :class:`~foundry.core.palette.Palette.Palette`.

    Attributes
    ----------
    type: PaletteType
        How the Palette should be loaded.
    """

    type: PaletteType

    class Config:
        # Allow storing the enum as a string
        use_enum_values = True

        # Enable cached property to be ignored by Pydantic
        arbitrary_types_allowed = True
        keep_untouched = (cached_property,)


class PydanticPaletteProtocol(Protocol):
    """
    A Pydantic palette that can convert to a regular palette.
    """

    @cached_property
    def palette(self) -> PaletteProtocol:
        ...


class PydanticColorsPalette(PydanticPalette):
    """
    A palette which generates itself from a series of indexes into the color palette.

    Attributes
    ----------
    color_indexes: list[int]
        A list of indexes into the color palette colors.
    color_palette: ColorPaletteCreator
        A color palette generator which provides the colors which are indexed.
    """

    color_indexes: list[int]
    color_palette: ColorPaletteCreator = ColorPaletteCreator(type="DEFAULT")

    @property
    def color_palette_(self) -> PydanticColorPaletteProtocol:
        """
        A simple wrapper around the true color palette to provide correct type hints.

        Returns
        -------
        ColorPaletteProtocol
            The color palette that is supplied.
        """
        return self.color_palette  # type: ignore

    @cached_property
    def palette(self) -> PaletteProtocol:
        """
        Provides the representation of a palette from the Pydantic version.

        Returns
        -------
        PaletteProtocol
            The corresponding palette.
        """
        return Palette.from_values(self.color_indexes, self.color_palette_.color_palette)


class PydanticROMAddressPalettePalette(PydanticPalette):
    """
    A palette which generates itself from an address in the ROM and the default color palette.

    Attributes
    ----------
    palette_address: int
        The index into the ROM to generate the palette from.
    """

    palette_address: int

    @cached_property
    def palette(self) -> PaletteProtocol:
        """
        Provides the representation of a palette from the ROM.

        Returns
        -------
        PaletteProtocol
            The corresponding palette.
        """
        return Palette.from_rom(self.palette_address)


class PaletteCreator(BaseModel):
    """
    A generator for a :class:`~foundry.core.palette.Palette.PaletteProtocol`.
    Creates the palette dynamically from its type attribute to provide it additional information.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        return super().__init_subclass__(**kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def generate_palette(cls, v: dict) -> PydanticPaletteProtocol:
        """
        The constructor for each specific palette.

        Parameters
        ----------
        v : dict
            The dictionary to create the palette from.

        Returns
        -------
        PydanticPaletteProtocol
            The created palette as defined by `v["type"]`

        Raises
        ------
        NotImplementedError
            If the constructor does not have a valid constructor for `v["type"]`.
        """

        type_ = PaletteType(v["type"])
        if type_ == PaletteType.from_colors:
            return PydanticColorsPalette(**v)
        if type_ == PaletteType.from_rom_address:
            return PydanticROMAddressPalettePalette(**v)
        raise NotImplementedError(f"There is no palette of type {type_}")

    @classmethod
    def validate(cls, v) -> PydanticPaletteProtocol:
        """
        Validates that the provided object is a valid palette.

        Parameters
        ----------
        v : dict
            The dictionary to create the palette from.

        Returns
        -------
        PydanticPaletteProtocol
            If validated, a palette will be created in accordance to `generate_palette`.

        Raises
        ------
        TypeError
            If a dictionary is not provided.
        TypeError
            If the dictionary does not contain the key `"type"`.
        TypeError
            If the type provided is not inside :class:`~foundry.core.palette.Color.PaletteType`.
        """
        if not isinstance(v, dict):
            raise TypeError("Dictionary required")
        if "type" not in v:
            raise TypeError("Must have a type")
        if not PaletteType.has_value(type_ := v["type"]):
            raise TypeError(f"{type_} is not a valid layout type")
        return cls.generate_palette(v)



from typing import List, Optional, Tuple
from warnings import warn

from PySide6.QtCore import QRect, QSize
from PySide6.QtGui import QColor, QImage, QPainter, Qt

from foundry.core.graphics_set.GraphicsSet import GraphicsSetProtocol
from foundry.core.palette.PaletteGroup import MutablePaletteGroup
from foundry.core.point.Point import Point, PointProtocol
from foundry.core.size.Size import Size, SizeProtocol
from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block, get_block
from foundry.game.gfx.objects.GeneratorObject import GeneratorObject
from foundry.game.gfx.objects.ObjectLike import (
    EXPANDS_BOTH,
    EXPANDS_HORIZ,
    EXPANDS_NOT,
    EXPANDS_VERT,
)
from foundry.game.ObjectDefinitions import EndType, GeneratorType, TilesetDefinition
from foundry.game.ObjectSet import ObjectSet
from foundry.smb3parse.objects.object_set import PLAINS_OBJECT_SET

SKY = 0
GROUND = 27

ENDING_STR = {
    EndType.UNIFORM: "Uniform",
    EndType.END_ON_TOP_OR_LEFT: "Top or Left",
    EndType.END_ON_BOTTOM_OR_RIGHT: "Bottom or Right",
    EndType.TWO_ENDS: "Top & Bottom/Left & Right",
}

ORIENTATION_TO_STR = {
    GeneratorType.HORIZONTAL: "Horizontal",
    GeneratorType.VERTICAL: "Vertical",
    GeneratorType.DIAG_DOWN_LEFT: "Diagonal ↙",
    GeneratorType.DESERT_PIPE_BOX: "Desert Pipe Box",
    GeneratorType.DIAG_DOWN_RIGHT: "Diagonal ↘",
    GeneratorType.DIAG_UP_RIGHT: "Diagonal ↗",
    GeneratorType.HORIZ_TO_GROUND: "Horizontal to the Ground",
    GeneratorType.HORIZONTAL_2: "Horizontal Alternative",
    GeneratorType.DIAG_WEIRD: "Diagonal Weird",  # up left?
    GeneratorType.SINGLE_BLOCK_OBJECT: "Single Block",
    GeneratorType.CENTERED: "Centered",
    GeneratorType.PYRAMID_TO_GROUND: "Pyramid to Ground",
    GeneratorType.PYRAMID_2: "Pyramid Alternative",
    GeneratorType.TO_THE_SKY: "To the Sky",
    GeneratorType.ENDING: "Ending",
}

# todo what is this, exactly?
ENDING_OBJECT_OFFSET = 0x1C8F9

# not all objects provide a block index for blank block
BLANK = -1

SCREEN_HEIGHT = 15
SCREEN_WIDTH = 16

MASK_COLOR = [0xFF, 0x33, 0xFF]


class LevelObject(GeneratorObject):
    def __init__(
        self,
        data: bytearray,
        object_set: int,
        palette_group: MutablePaletteGroup,
        graphics_set: GraphicsSetProtocol,
        objects_ref: List["LevelObject"],
        is_vertical: bool,
        index: int,
        size_minimal: bool = False,
    ):
        self.object_set = ObjectSet(object_set)

        self.graphics_set = graphics_set
        self._position = Point(0, 0)
        self._ignore_rendered_position = False

        self.palette_group = tuple(tuple(c for c in pal) for pal in palette_group)

        self.index_in_level = index
        self.objects_ref = objects_ref
        self.vertical_level = is_vertical

        self.data = data

        self.selected = False

        self.size_minimal = size_minimal

        if self.size_minimal:
            self.ground_level = 0
        else:
            self.ground_level = GROUND

        self.render()

    @property
    def domain(self) -> int:
        return (self.data[0] & 0b1110_0000) >> 5

    @property
    def orientation(self) -> GeneratorType:
        return GeneratorType(self.definition.orientation)

    @property
    def ending(self) -> EndType:
        return EndType(self.definition.ending)

    @property
    def name(self) -> str:
        return self.definition.description

    @property
    def blocks(self) -> list[int]:
        return self.definition.blocks

    @property
    def size(self) -> int:
        return self.definition.size

    @property
    def is_4byte(self) -> bool:
        return self.size == 4

    @property
    def tsa_data(self) -> bytearray:
        return ROM.get_tsa_data(self.object_set.number)

    @property
    def is_single_block(self) -> bool:
        return self.obj_index <= 0x0F

    @property
    def type(self) -> int:
        domain_offset = self.domain * 0x1F

        if self.is_single_block:
            return self.obj_index + domain_offset
        else:
            return (self.obj_index >> 4) + domain_offset + 16 - 1

    @property
    def definition(self) -> TilesetDefinition:
        return self.object_set.get_definition_of(self.type)

    @property
    def obj_index(self) -> int:
        return self.data[2]

    @obj_index.setter
    def obj_index(self, value: int):
        self.data[2] = value

    @property
    def object_info(self):
        return self.object_set.number, self.domain, self.obj_index

    @property
    def length(self) -> int:
        if self.is_single_block:
            return 1
        if self.size == 3:
            return self.obj_index & 0x0F
        else:
            try:
                return self.data[3]
            except IndexError:
                return 0

    @length.setter
    def length(self, value: int):
        if not self.is_single_block:
            if not self.is_4byte:
                index = self.obj_index
                index &= 0xF0
                index |= value & 0x0F
                self.obj_index = index
            else:
                try:
                    self.data[3] = value
                except IndexError:
                    self.data.append(value)

    @property
    def secondary_length(self) -> int:
        if self.size == 3:
            return 1
        else:
            return self.obj_index & 0x0F

    @secondary_length.setter
    def secondary_length(self, value: int):
        if self.size >= 4:
            index = self.obj_index
            index &= 0xF0
            index |= value & 0x0F
            self.obj_index = index

    def render(self):
        self._render()

    def _render(self):
        try:
            self.index_in_level = self.objects_ref.index(self)
        except ValueError:
            # the object has not been added yet, so stick with the one given in the constructor
            pass

        blocks_to_draw = []

        if self.orientation == GeneratorType.TO_THE_SKY:
            for _ in range(self.position.y):
                blocks_to_draw.extend(self.blocks[0 : self.scale.width])

            blocks_to_draw.extend(self.blocks[-self.scale.width :])

        elif self.orientation == GeneratorType.DESERT_PIPE_BOX:
            # segments are the horizontal sections, which are 8 blocks long
            # two of those are drawn per length bit
            # rows are the 4 block high rows Mario can walk in

            is_pipe_box_type_b = self.obj_index // 0x10 == 4

            rows_per_box = self.scale.height
            lines_per_row = 4

            segment_width = self.scale.width
            segments = (self.length + 1) * 2

            for row_number in range(rows_per_box):
                for line in range(lines_per_row):
                    if is_pipe_box_type_b and row_number > 0 and line == 0:
                        # in pipebox type b we do not repeat the horizontal beams
                        line += 1

                    start = line * segment_width
                    stop = start + segment_width

                    for segment_number in range(segments):
                        blocks_to_draw.extend(self.blocks[start:stop])

            if is_pipe_box_type_b:
                # draw another open row
                start = segment_width
            else:
                # draw the first row again to close the box
                start = 0

            stop = start + segment_width

            for segment_number in range(segments):
                blocks_to_draw.extend(self.blocks[start:stop])

            # every line repeats the last block again for some reason
            for end_of_line in range(len(blocks_to_draw), 0, -(self.rendered_size.width - 1)):
                blocks_to_draw.insert(end_of_line, blocks_to_draw[end_of_line - 1])

        elif self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            if self.ending == EndType.UNIFORM:
                left = [BLANK]
                right = [BLANK]
                slopes = self.blocks

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                if self.orientation in [GeneratorType.DIAG_DOWN_RIGHT, GeneratorType.DIAG_UP_RIGHT]:
                    fill_block = self.blocks[0:1]
                    slopes = self.blocks[1:]

                    left = fill_block
                    right = [BLANK]
                elif self.orientation == GeneratorType.DIAG_DOWN_LEFT:
                    fill_block = self.blocks[-1:]
                    slopes = self.blocks[0:-1]

                    right = fill_block
                    left = [BLANK]

                else:
                    fill_block = self.blocks[0:1]
                    slopes = self.blocks[1:]

                    right = [BLANK]
                    left = fill_block

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                fill_block = self.blocks[-1:]
                slopes = self.blocks[0:-1]

                left = [BLANK]
                right = fill_block
            else:
                # todo other two ends not used with diagonals?
                warn(f"{self.name} was not rendered.", RuntimeWarning)
                self.rendered_blocks = []
                return

            rows = []

            if self.scale.height > self.scale.width:
                slope_width = self.scale.width
            else:
                slope_width = len(slopes)

            for y in range(self.rendered_size.height):
                amount_right = (y // self.scale.height) * slope_width
                amount_left = self.rendered_size.width - slope_width - amount_right

                offset = y % self.scale.height

                rows.append(amount_left * left + slopes[offset : offset + slope_width] + amount_right * right)

            if self.orientation in [GeneratorType.DIAG_UP_RIGHT]:
                for row in rows:
                    row.reverse()

            if self.orientation in [GeneratorType.DIAG_DOWN_RIGHT, GeneratorType.DIAG_UP_RIGHT]:
                if not self.scale.height > self.scale.width:
                    rows.reverse()

            if self.orientation == GeneratorType.DIAG_DOWN_RIGHT and self.scale.height > self.scale.width:
                # special case for 60 degree platform wire down right
                for row in rows:
                    row.reverse()

            for row in rows:
                blocks_to_draw.extend(row)

        elif self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            # since pyramids grow horizontally in both directions when extending
            # we need to check for new ground every time it grows
            rendered_size = self.rendered_size

            blank = self.blocks[0]
            left_slope = self.blocks[1]
            left_fill = self.blocks[2]
            right_fill = self.blocks[3]
            right_slope = self.blocks[4]

            for y in range(rendered_size.height):
                blank_blocks = (rendered_size.width // 2) - (y + 1)
                middle_blocks = y  # times two

                blocks_to_draw.extend(blank_blocks * [blank])

                blocks_to_draw.append(left_slope)
                blocks_to_draw.extend(middle_blocks * [left_fill] + middle_blocks * [right_fill])
                blocks_to_draw.append(right_slope)

                blocks_to_draw.extend(blank_blocks * [blank])

        elif self.orientation == GeneratorType.ENDING:
            page_width = 16
            page_limit = page_width - self.position.x % page_width

            for y in range(SKY, GROUND - 1):
                blocks_to_draw.append(self.blocks[0])
                blocks_to_draw.extend([self.blocks[1]] * (self.rendered_size.width))

            # todo magic number
            # ending graphics
            rom_offset = ENDING_OBJECT_OFFSET + self.object_set.get_ending_offset() * 0x60

            rom = ROM()

            ending_graphic_height = 6
            floor_height = 1

            y_offset = GROUND - floor_height - ending_graphic_height

            for y in range(ending_graphic_height):
                for x in range(page_width):
                    block_index = rom.get_byte(rom_offset + y * page_width + x - 1)

                    block_position = (y_offset + y) * (self.rendered_size.width + 1) + x + page_limit + 1
                    blocks_to_draw[block_position] = block_index

            # the ending object is seemingly always 1 block too wide (going into the next screen)
            for end_of_line in range(len(blocks_to_draw) - 1, 0, -(self.rendered_size.width + 1)):
                del blocks_to_draw[end_of_line]

        elif self.orientation == GeneratorType.VERTICAL:
            if self.ending == EndType.UNIFORM:
                for _ in range(self.length + 1):
                    for y in range(self.scale.height):
                        for x in range(self.rendered_size.width):
                            blocks_to_draw.append(self.blocks[y * self.scale.height + x % self.scale.width])

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                # in case the drawn object is smaller than its actual size
                for y in range(min(self.scale.height, self.rendered_size.height)):
                    offset = y * self.scale.width
                    blocks_to_draw.extend(self.blocks[offset : offset + self.scale.width])

                additional_rows = self.rendered_size.height - self.scale.height

                # assume only the last row needs to repeat
                # todo true for giant blocks?
                if additional_rows > 0:
                    last_row = self.blocks[-self.scale.width :]

                    for _ in range(additional_rows):
                        blocks_to_draw.extend(last_row)

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                additional_rows = self.rendered_size.height - self.scale.height

                # assume only the first row needs to repeat
                # todo true for giant blocks?
                if additional_rows > 0:
                    last_row = self.blocks[0 : self.scale.width]

                    for _ in range(additional_rows):
                        blocks_to_draw.extend(last_row)

                # in case the drawn object is smaller than its actual size
                for y in range(min(self.scale.height, self.rendered_size.height)):
                    offset = y * self.scale.width
                    blocks_to_draw.extend(self.blocks[offset : offset + self.scale.width])

            elif self.ending == EndType.TWO_ENDS:
                # object exists on ships
                top_row = self.blocks[0 : self.scale.width]
                bottom_row = self.blocks[-self.scale.width :]

                blocks_to_draw.extend(top_row)

                additional_rows = self.rendered_size.height - 2

                # repeat second to last row
                if additional_rows > 0:
                    for _ in range(additional_rows):
                        blocks_to_draw.extend(self.blocks[-2 * self.scale.width : -self.scale.width])

                if self.rendered_size.height > 1:
                    blocks_to_draw.extend(bottom_row)

        elif self.orientation in [GeneratorType.HORIZONTAL, GeneratorType.HORIZ_TO_GROUND, GeneratorType.HORIZONTAL_2]:
            if self.ending == EndType.UNIFORM and not self.is_4byte:
                for y in range(self.rendered_size.height):
                    if self.is_single_block:
                        blocks_to_draw.extend(self.blocks[: self.scale.width])
                    else:
                        blocks_to_draw.extend(
                            self.blocks[y * self.scale.width : (y + 1) * self.scale.width] * (self.length + 1)
                        )

            elif self.ending == EndType.UNIFORM and self.is_4byte:
                # 4 byte objects
                top = self.blocks[0:1]
                bottom = self.blocks[-1:]

                if self.orientation == GeneratorType.HORIZONTAL_2:
                    for _ in range(0, self.rendered_size.height - 1):
                        blocks_to_draw.extend(self.rendered_size.width * top)

                    blocks_to_draw.extend(self.rendered_size.width * bottom)
                else:
                    blocks_to_draw.extend(self.rendered_size.width * top)

                    for _ in range(1, self.rendered_size.height):
                        blocks_to_draw.extend(self.rendered_size.width * bottom)

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                for y in range(self.rendered_size.height):
                    offset = y * self.scale.width

                    blocks_to_draw.append(self.blocks[offset])

                    for x in range(1, self.rendered_size.width):
                        blocks_to_draw.append(self.blocks[offset + 1])

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                for y in range(self.rendered_size.height):
                    offset = y * self.scale.width

                    for x in range(self.rendered_size.width - 1):
                        blocks_to_draw.append(self.blocks[offset])

                    blocks_to_draw.append(self.blocks[offset + self.scale.width - 1])

            elif self.ending == EndType.TWO_ENDS:
                if self.scale.width > len(self.blocks):
                    raise ValueError(f"{self} does not provide enough blocks to fill a row.")
                else:
                    start = 0
                    end = self.scale.width

                for y in range(self.scale.height):
                    new_start = y * self.scale.width
                    new_end = (y + 1) * self.scale.width

                    if new_end > len(self.blocks):
                        # repeat the last line of blocks to fill the object
                        pass
                    else:
                        start = new_start
                        end = new_end

                    left, *middle, right = self.blocks[start:end]

                    blocks_to_draw.append(left)
                    blocks_to_draw.extend(middle * (self.rendered_size.width - 2))
                    blocks_to_draw.append(right)

                if not len(blocks_to_draw) % self.scale.height == 0:
                    warn(f"Blocks to draw are not divisible by height. {self}", RuntimeWarning)

                new_width = int(len(blocks_to_draw) / self.scale.height)

                top_row = blocks_to_draw[0:new_width]
                middle_blocks = blocks_to_draw[new_width : new_width * 2]
                bottom_row = blocks_to_draw[-new_width:]

                blocks_to_draw = top_row

                for y in range(1, self.rendered_size.height - 1):
                    blocks_to_draw.extend(middle_blocks)

                if self.rendered_size.height > 1:
                    blocks_to_draw.extend(bottom_row)
        else:
            if not self.orientation == GeneratorType.SINGLE_BLOCK_OBJECT:
                warn(f"Didn't render {self.name}", RuntimeWarning)
                # breakpoint()

            if self.name.lower() == "black boss room background":
                blocks_to_draw = SCREEN_WIDTH * SCREEN_HEIGHT * [self.blocks[0]]

        # for not yet implemented objects and single block objects
        if blocks_to_draw:
            self.rendered_blocks = blocks_to_draw
        else:
            self.rendered_blocks = self.blocks

        self.rect = QRect(
            self.rendered_position.x, self.rendered_position.y, self.rendered_size.width, self.rendered_size.height
        )

    def draw(self, painter: QPainter, block_length, transparent, blocks: Optional[list[Block]] = None):
        size = self.rendered_size
        size.width = max(size.width, 1)

        for index, block_index in enumerate(self.rendered_blocks):
            if block_index == BLANK:
                continue

            x = self.rendered_position.x + index % size.width
            y = self.rendered_position.y + index // size.width

            self._draw_block(painter, block_index, x, y, block_length, transparent, blocks=blocks)

    def _draw_block(
        self, painter: QPainter, block_index, x, y, block_length, transparent, blocks: Optional[list[Block]] = None
    ):
        if blocks is not None:
            block = blocks[block_index if block_index <= 0xFF else ROM().get_byte(block_index)]
        else:
            block = get_block(
                block_index,
                self.palette_group,
                self.graphics_set,
                bytes(self.tsa_data),
            )

        block.draw(
            painter,
            x * block_length,
            y * block_length,
            block_length=block_length,
            selected=self.selected,
            transparent=transparent,
        )

    def move_by(self, dx: int, dy: int):
        self.position = Point(self.position.x + dx, self.position.y + dy)

    @property
    def position(self) -> PointProtocol:
        y = self.data[0] & 0b0001_1111
        x = self.data[1]

        if self.vertical_level:
            offset = (x // SCREEN_WIDTH) * SCREEN_HEIGHT

            y += offset
            x %= SCREEN_WIDTH

        return Point(x, y)

    @position.setter
    def position(self, position: PointProtocol) -> None:
        x, y = position.x, position.y

        # todo also check for the upper bounds
        x = max(0, x)

        if self.vertical_level:
            # todo from vertical to non-vertical is bugged, because it
            # seems like you can't convert the coordinates 1:1
            # there seems to be ambiguity

            offset = y // SCREEN_HEIGHT

            x += offset * SCREEN_WIDTH
            y %= SCREEN_HEIGHT

        self.data[0] = (self.data[0] & 0b1110_0000) + y
        self.data[1] = x

        self._render()

    @property
    def rendered_position(self) -> PointProtocol:
        if self._ignore_rendered_position:
            return Point(0, 0)
        elif self.orientation == GeneratorType.TO_THE_SKY:
            return Point(self.position.x, SKY)
        elif self.orientation in [GeneratorType.DIAG_UP_RIGHT]:
            return Point(self.position.x, self.position.y - self.rendered_size.height + 1)
        elif self.orientation in [GeneratorType.DIAG_DOWN_LEFT]:
            if self.object_set.number == 3 or self.object_set.number == 14:  # Sky or Hilly tileset
                return Point(self.position.x - (self.rendered_size.width - self.scale.width + 1), self.position.y)
            else:
                return Point(self.position.x - (self.rendered_size.width - self.scale.width), self.position.y)

        elif self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            return Point(self.position.x - (self.rendered_size.width // 2) + 1, self.position.y)
        elif self.name.lower() == "black boss room background":
            return Point(self.position.x // SCREEN_WIDTH * SCREEN_WIDTH, 0)
        return self.position

    @property
    def scale(self) -> SizeProtocol:
        return Size(self.definition.bmp_width, self.definition.bmp_height)

    @property
    def rendered_size(self) -> SizeProtocol:
        if self.orientation == GeneratorType.TO_THE_SKY:
            return Size(self.scale.width, self.position.y + self.scale.height - 1)
        elif self.orientation == GeneratorType.DESERT_PIPE_BOX:
            segments = (self.length + 1) * 2
            return Size(segments * self.scale.width + 1, 4 * self.scale.height)
        elif self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            if self.ending == EndType.UNIFORM:
                return Size((self.length + 1) * self.scale.width, (self.length + 1) * self.scale.height)
            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                return Size((self.length + 1) * (self.scale.width - 1), (self.length + 1))
            else:
                return Size((self.length + 1) * (self.scale.width - 1), (self.length + 1) * self.scale.height)
        elif self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            size = Size(1, 1)
            for y in range(self.position.y, self.ground_level):
                size = Size(2 * (y - self.position.y), (y - self.position.y))
                bottom_row = QRect(self.position.x, y, size.width, 1)
                if any(
                    [
                        bottom_row.intersects(obj.get_rect()) and y == obj.get_rect().top()
                        for obj in self.objects_ref[0 : self.index_in_level]
                    ]
                ):
                    break
            return size
        elif self.orientation == GeneratorType.ENDING:
            page_width = 16
            page_limit = page_width - self.position.x % page_width
            return Size(page_width + page_limit, (GROUND - 1) - SKY)
        elif self.orientation == GeneratorType.VERTICAL:
            size = Size(self.scale.width, self.length + 1)

            if self.ending == EndType.UNIFORM:
                if self.is_4byte:
                    # there is one VERTICAL 4-byte object: Vertically oriented X-blocks
                    # the width is the primary expansion
                    size.width = (self.obj_index & 0x0F) + 1

                # adjust height for giant blocks, so that the rect is correct
                size.height *= self.scale.height
            return size
        elif self.orientation in [GeneratorType.HORIZONTAL, GeneratorType.HORIZ_TO_GROUND, GeneratorType.HORIZONTAL_2]:
            size = Size(self.length + 1, self.scale.height)

            downwards_extending_vine = 1, 0, 0x06
            wooden_sky_pole = 4, 0, 0x04

            if self.object_info in [downwards_extending_vine, wooden_sky_pole]:
                size.width -= 1
            if self.orientation == GeneratorType.HORIZ_TO_GROUND:
                # to the ground only, until it hits something
                for y in range(self.position.y, self.ground_level):
                    bottom_row = QRect(self.position.x, y, size.width, 1)

                    if any(
                        [
                            bottom_row.intersects(obj.get_rect()) and y == obj.get_rect().top()
                            for obj in self.objects_ref[0 : self.index_in_level]
                        ]
                    ):
                        size.height = y - self.position.y
                        break
                else:
                    # nothing underneath this object, extend to the ground
                    size.height = self.ground_level - self.position.y

                if self.is_single_block:
                    size.width = self.length

                size.height = max(min(self.scale.height, 2), size.height)

            elif self.orientation == GeneratorType.HORIZONTAL_2 and self.ending == EndType.TWO_ENDS:
                # floating platforms seem to just be one shorter for some reason
                size.width -= 1
            else:
                size.height = self.scale.height * (self.secondary_length)

            if self.ending == EndType.UNIFORM and not self.is_4byte:
                size.width *= self.scale.width  # in case of giant blocks
            elif self.ending == EndType.UNIFORM and self.is_4byte:
                size.height = self.scale.height + self.secondary_length

                # ceilings are one shorter than normal
                if self.scale.height > self.scale.width:
                    size.height -= 1

            elif self.ending == EndType.TWO_ENDS:
                if self.orientation == GeneratorType.HORIZONTAL and self.is_4byte:
                    # flat ground objects have an artificial limit of 2 lines
                    if (
                        self.object_set.number == PLAINS_OBJECT_SET
                        and self.domain == 0
                        and self.obj_index in range(0xC0, 0xE0)
                    ):
                        size.height = min(2, self.secondary_length + 1)
                    else:
                        size.height = self.secondary_length + 1
            return size
        elif self.name.lower() == "black boss room background":
            return Size(SCREEN_WIDTH, SCREEN_HEIGHT)
        return self.scale

    @property
    def horizontally_expands(self) -> bool:
        return bool(self.expands() & EXPANDS_HORIZ)

    @property
    def vertically_expands(self) -> bool:
        return bool(self.expands() & EXPANDS_VERT)

    def expands(self):
        expands = EXPANDS_NOT

        if self.is_single_block:
            return expands

        if self.is_4byte:
            expands |= EXPANDS_BOTH

        elif self.orientation in [
            GeneratorType.HORIZONTAL,
            GeneratorType.HORIZONTAL_2,
            GeneratorType.HORIZ_TO_GROUND,
        ] or self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            expands |= EXPANDS_HORIZ

        elif self.orientation in [GeneratorType.VERTICAL, GeneratorType.DIAG_WEIRD]:
            expands |= EXPANDS_VERT

        return expands

    def primary_expansion(self):
        if self.orientation in [
            GeneratorType.HORIZONTAL,
            GeneratorType.HORIZONTAL_2,
            GeneratorType.HORIZ_TO_GROUND,
        ] or self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            if self.is_4byte:
                return EXPANDS_VERT
            else:
                return EXPANDS_HORIZ
        elif self.orientation in [GeneratorType.VERTICAL]:
            if self.is_4byte:
                return EXPANDS_HORIZ
            else:
                return EXPANDS_VERT
        else:
            return EXPANDS_BOTH

    def __contains__(self, item: Tuple[int, int]) -> bool:
        x, y = item

        return self.point_in(x, y)

    def point_in(self, x: int, y: int) -> bool:
        return self.rect.contains(x, y)

    def get_status_info(self) -> List[tuple]:
        return [
            ("x", self.rendered_position.x),
            ("y", self.rendered_position.y),
            ("Width", self.rendered_size.width),
            ("Height", self.rendered_size.height),
            ("Orientation", ORIENTATION_TO_STR[self.orientation]),
            ("Ending", ENDING_STR[self.ending]),
        ]

    def display_size(self, zoom_factor: int = 1):
        return (
            QSize(self.rendered_size.width * Block.SIDE_LENGTH, self.rendered_size.height * Block.SIDE_LENGTH)
            * zoom_factor
        )

    def as_image(self) -> QImage:
        self._ignore_rendered_position = True

        image = QImage(
            QSize(self.rendered_size.width * Block.SIDE_LENGTH, self.rendered_size.height * Block.SIDE_LENGTH),
            QImage.Format_RGB888,
        )

        bg_color = QColor(*MASK_COLOR).rgb()

        image.fill(bg_color)
        mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
        image.setAlphaChannel(mask)

        painter = QPainter(image)

        self.draw(painter, Block.SIDE_LENGTH, True)

        self._ignore_rendered_position = False

        return image

    def to_bytes(self) -> bytearray:
        data = bytearray()

        if self.vertical_level:
            # todo from vertical to non-vertical is bugged, because it
            # seems like you can't convert the coordinates 1:1
            # there seems to be ambiguity

            offset = self.position.y // SCREEN_HEIGHT

            x_position = self.position.x + offset * SCREEN_WIDTH
            y_position = self.position.y % SCREEN_HEIGHT
        else:
            x_position = self.position.x
            y_position = self.position.y

        if self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            x_position = self.rendered_position.x - 1 + self.rendered_size.width // 2

        data.append((self.domain << 5) | y_position)
        data.append(x_position)

        data.append(self.obj_index)

        if self.is_4byte:
            data.append(self.length)

        return data

    def __repr__(self) -> str:
        return f"LevelObject {self.name} at {self.position.x}, {self.position.y}"

    def __eq__(self, other):
        if not isinstance(other, LevelObject):
            return False
        else:
            return self.to_bytes() == other.to_bytes()

    def __lt__(self, other):
        return self.index_in_level < other.index_in_level



from itertools import chain
from typing import Protocol

from attr import attrs
from pydantic import BaseModel

from foundry.core.graphics_page.GraphicsPage import (
    GraphicsPageProtocol,
    PydanticGraphicsPage,
)
from foundry.core.graphics_set.util import get_graphics_pages_from_tileset


class GraphicsSetProtocol(Protocol):
    """
    A representation of a series of graphical pages inside the ROM.

    Attributes
    ----------
    pages: tuple[GraphicsPageProtocol, ...]
        The pages that compose the graphical set.
    """

    pages: tuple[GraphicsPageProtocol, ...]

    def __bytes__(self) -> bytes:
        ...


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class GraphicsSet:
    """
    A representation of a series of graphical pages inside the ROM, that uses ``attrs`` to create
    a basic implementation.

    Attributes
    ----------
    pages: tuple[GraphicsPageProtocol, ...]
        The pages that compose the graphical set.
    """

    pages: tuple[GraphicsPageProtocol, ...]

    def __bytes__(self) -> bytes:
        return bytes(chain.from_iterable([bytes(page) for page in self.pages]))

    @classmethod
    def from_tileset(cls, index: int):
        cls.number = index
        return cls(get_graphics_pages_from_tileset(index))


class PydanticGraphicsPage(BaseModel):
    """
    A JSON model of a generic GraphicsSet through Pydantic.

    Attributes
    ----------
    pages: list[PydanticGraphicsPage]
        The pages that compose the graphical set.
    """

    pages: list[PydanticGraphicsPage]

    def to_graphics_set(self) -> GraphicsSetProtocol:
        return GraphicsSet(tuple(page.to_graphics_page() for page in self.pages))



from functools import cache

from foundry.core.palette import (
    PALETTE_BASE_ADDRESS,
    PALETTE_OFFSET_LIST,
    PALETTE_OFFSET_SIZE,
)
from foundry.game.File import ROM


@cache
def get_internal_palette_offset(tileset: int) -> int:
    """
    Provides the absolute internal point of the palette group offset from ROM.

    Parameters
    ----------
    tileset : int
        The tileset to find the absolute internal point of.

    Returns
    -------
    int
        The absolute internal point of the tileset's palette group.
    """
    return PALETTE_BASE_ADDRESS + ROM().little_endian(PALETTE_OFFSET_LIST + (tileset * PALETTE_OFFSET_SIZE))



from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Protocol, Sequence

from attr import attrs
from pydantic import BaseModel, FilePath

from foundry.core.palette import PALETTE_FILE_COLOR_OFFSET, PALETTE_FILE_PATH
from foundry.core.palette.Color import Color, ColorProtocol, PydanticColor


class ColorPaletteProtocol(Protocol):
    colors: Sequence[ColorProtocol]


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class ColorPalette:
    """
    A representation of a series of colors.
    """

    colors: tuple[ColorProtocol, ...]


class ColorPaletteType(str, Enum):
    """
    A declaration of the color palettes possible to be created through
    `JSON <https://en.wikipedia.org/wiki/JSON>`_ and
    `Pydantic <https://pydantic-docs.helpmanual.io/>`_.
    """

    default = "DEFAULT"
    colors = "COLORS"
    palette_file = "PALETTE FILE"

    @classmethod
    def has_value(cls, value):
        """
        A convenience method to quickly determine if a value is a valid enumeration.

        Parameters
        ----------
        value : str
            The value to check against the enumeration.

        Returns
        -------
        bool
            If the value is inside the enumeration.
        """

        return value in cls._value2member_map_


class PydanticColorPalette(BaseModel):
    """
    A generic representation of :class:`~foundry.core.palette.ColorPalette.ColorPalette`.

    Attributes
    ----------
    type: ColorPaletteType
        How the ColorPalette should be loaded.
    """

    type: ColorPaletteType

    class Config:
        # Allow storing the enum as a string
        use_enum_values = True

        # Enable cached property to be ignored by Pydantic
        arbitrary_types_allowed = True
        keep_untouched = (cached_property,)


class PydanticColorPaletteProtocol(Protocol):
    """
    A Pydantic color palette that can convert to a regular color palette.
    """

    @cached_property
    def color_palette(self) -> ColorPaletteProtocol:
        ...


class PydanticColorsColorPalette(PydanticColorPalette):
    """
    A color palette which generates itself from a series of colors.

    Attributes
    ----------
    colors: list[PydanticColor]
        A list of colors that compose the color palette.
    """

    colors: list[PydanticColor]

    @cached_property
    def color_palette(self) -> ColorPaletteProtocol:
        """
        Provides the representation of a color palette from the Pydantic version.

        Returns
        -------
        ColorPaletteProtocol
            The corresponding color palette.
        """
        return ColorPalette(tuple(color.color for color in self.colors))


def palette_file_to_color_palette(path: Path) -> ColorPaletteProtocol:
    """
    Generates a color palette from a palette file.

    Parameters
    ----------
    path : Path
        The path to the palette file.

    Returns
    -------
    ColorPaletteProtocol
        The color palette that represents the palette file.
    """
    with open(path, "rb") as f:
        data = f.read()
    return ColorPalette(
        tuple(Color(data[i], data[i + 1], data[i + 2]) for i in range(PALETTE_FILE_COLOR_OFFSET, len(data), 4))
    )


class PydanticPaletteFileColorPalette(PydanticColorPalette):
    """
    A color palette which generates itself from another file.

    Attributes
    ----------
    path: FilePath
        A palette file to be converted to a color palette.
    """

    path: FilePath

    @cached_property
    def color_palette(self) -> ColorPaletteProtocol:
        """
        Provides the representation of a color palette from the Pydantic version.

        Returns
        -------
        ColorPaletteProtocol
            The corresponding color palette.
        """
        return palette_file_to_color_palette(self.path)


class PydanticDefaultColorPalette(PydanticColorPalette):
    """
    A color palette which represents the NES colors inside the game.
    """

    @cached_property
    def color_palette(self) -> ColorPaletteProtocol:
        """
        Provides the representation of a color palette from the Pydantic version.

        Returns
        -------
        ColorPaletteProtocol
            The corresponding color palette.
        """
        return PydanticPaletteFileColorPalette(type="PALETTE FILE", path=PALETTE_FILE_PATH).color_palette


class ColorPaletteCreator(BaseModel):
    """
    A generator for a :class:`~foundry.core.palette.ColorPalette.ColorPaletteProtocol`.
    Creates the color palette dynamically from its type attribute to provide it additional information.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        return super().__init_subclass__(**kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def as_default(cls) -> PydanticColorPaletteProtocol:
        return PydanticDefaultColorPalette(type="DEFAULT")

    @classmethod
    def generate_color_palette(cls, v: dict) -> PydanticColorPaletteProtocol:
        """
        The constructor for each specific color palette.

        Parameters
        ----------
        v : dict
            The dictionary to create the color palette from.

        Returns
        -------
        PydanticColorPaletteProtocol
            The created layout as defined by `v["type"]`

        Raises
        ------
        NotImplementedError
            If the constructor does not have a valid constructor for `v["type"]`.
        """

        type_ = ColorPaletteType(v["type"])
        if type_ == ColorPaletteType.default:
            return PydanticDefaultColorPalette(**v)
        if type_ == ColorPaletteType.colors:
            return PydanticColorsColorPalette(**v)
        if type_ == ColorPaletteType.palette_file:
            return PydanticPaletteFileColorPalette(**v)
        raise NotImplementedError(f"There is no color palette of type {type_}")

    @classmethod
    def validate(cls, v) -> PydanticColorPaletteProtocol:
        """
        Validates that the provided object is a valid color palette.

        Parameters
        ----------
        v : dict
            The dictionary to create the color palette from.

        Returns
        -------
        PydanticColorPaletteProtocol
            If validated, a color palette will be created in accordance to `generate_color_palette`.

        Raises
        ------
        TypeError
            If a dictionary is not provided.
        TypeError
            If the dictionary does not contain the key `"type"`.
        TypeError
            If the type provided is not inside :class:`~foundry.core.palette.ColorPalette.ColorPaletteType`.
        """
        if not isinstance(v, dict):
            raise TypeError("Dictionary required")
        if "type" not in v:
            raise TypeError("Must have a type")
        if not ColorPaletteType.has_value(type_ := v["type"]):
            raise TypeError(f"{type_} is not a valid layout type")
        return cls.generate_color_palette(v)
